"""Base classes for basic components.
Each package-based components such as Trainer or Grader inherit classes from this module."""

from abc import ABCMeta, abstractmethod
from typing import TypeVar, Callable, Sequence, Union, Any, Hashable
from .course import CourseBook
from ..utils import load_class
import copy
import warnings
from omegaconf import DictConfig

T_co = TypeVar("T_co", covariant=True)
B = TypeVar("B", bound=CourseBook)


class ConfigurationError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class Warning:
    def __init__(self, logger) -> None:
        self.logger = logger

    def __call__(self, message, category, filename, lineno, file=None, line=None):
        self.logger.warning(
            warnings.formatwarning(message, category, filename, lineno, line)
        )


class Learner(metaclass=ABCMeta):
    r"""Base class for all learners
    It is basically wrapper class for model such as module in pytorch or estimator in scikit-learn.
    This class is meant for just simple training and inference features, not for some other things like datastream manipulation.
    It is because the model could be used for inference only task in the future updates."""

    @abstractmethod
    def predict(self, batch):
        raise NotImplementedError(
            f"'predict' method of {self.__class__.__name__} class is not implemented."
        )

    @abstractmethod
    def fit(self, batch):
        pass

    @abstractmethod
    def resume_model(self):
        ...


L = TypeVar("L", bound=Learner)


class Grader:
    class EvaluationResult:
        def __init__(self) -> None:
            self.__result = {}

        def done(self, data: Any, label: Hashable = None):
            if isinstance(data, dict):
                for key, value in data.items():
                    self.__result[key] = value
            else:
                self.__result[label] = data

        def as_dict(self) -> dict:
            return copy.deepcopy(self.__result)

        def __repr__(self) -> str:
            return "\n".join(
                [f"{label:20} : {res}" for label, res in self.__result.items()]
            )

    def __init__(
        self,
        conf: DictConfig,
        methods: Union[str, Callable, Sequence[Union[str, Callable]]] = None,
        options: list[dict] = None,
        logger=None,
        course: B = None,
        learner: L = None,
    ) -> None:
        self.methods = []
        self.conf = conf
        if conf.scheme.metrics:
            self.append_metric(conf.scheme.metrics)
        self.course = course
        self.learner = learner
        self.append_metric(methods)
        self.log = logger
        self.options = options

    def append_metric(self, methods: Union[Callable, Sequence[Callable]]):
        if callable(methods):
            self.methods.append(methods)
        elif methods is not None:
            for method in methods:
                if callable(method):
                    self.methods.append(method)
                else:
                    if isinstance(method, str):
                        self.methods.append(load_class(_class=method))
                    elif isinstance(method, DictConfig):
                        self.methods.append(
                            load_class(
                                method.path if "path" in method else None, method.cls
                            )
                        )
                    else:
                        raise ValueError("Invalid metric")

    def is_enabled(self) -> bool:
        return len(self.methods) > 0

    def compute(
        self, prediction: T_co, ground_truth: T_co, options: list[dict] = None
    ) -> dict:
        result = {}
        if options is None:
            options = self.options

        for i, method in enumerate(self.methods):
            result[method.__name__] = (
                method(prediction, ground_truth, options[i])
                if options and options[i]
                else method(prediction, ground_truth)
            )

        return result

    def __call__(self, prediction: T_co, ground_truth: T_co, options: list[dict]):
        self.compute(prediction, ground_truth)


class RemoteGrader:
    """Spawn a separated evalaution process"""

    def __init__(self) -> None:
        raise NotImplementedError()

    def run() -> None:
        pass


class Trainer:
    r"""Base class for Trainer

    Trainer provides logging and saving checkpoints, furthermore lessen the repetitive parts in usual machine learning codes.
    BaseTrainer provides fit, eval, and predict features even it not specifying compatibility with libraries such as PyTorch or Scikit-learn.
    It can work with some of simple models, thus you can use this standalone. But for supported libraries, it would be recommended to use based class for the task.
    Remember, Trainers are designed to do only one task at a time, which means for multiple tasks you are required to initiate the same number of trainers.
    This is beacause for distributed learning in PyTorch, it would complicate the whole project if the trainer manages multiple tasks by alone.

    Arguments:
        :param course: dataset for the task. It should not be modified while on doing tasks.
        :param learner: learner(model) for the task. recommend to use same based class e.g. TorchLearner for TorchTrainer.
        :param conf: Trainer configuration structured by omegaconf DictConfig.
        :param grader: Evalutor from eniat package.
        :param logger: Logger from eniat package. required.
    """

    def __init__(
        self, course: B = None, learner: L = None, conf=None, grader=None, logger=None
    ) -> None:
        self.course = course
        self.learner = learner
        self.conf = conf
        self.grader = grader
        self.log = logger

    @abstractmethod
    def fit(self, silent: bool = False, eval: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def predict(self):
        raise NotImplementedError()



"""
subpacakge 'core' is for pacakge-independent components.

Basic components including Trainer, Grader, Learner
provided as abstract classes. And Course and CourseBook
that works as data handler also provided here as it is
package-independent.
"""

from .base import Learner, Grader, RemoteGrader, Trainer, Warning, ConfigurationError
from .course import Course, CourseBook, get_course_instance, batch_load

__all__ = (
    Learner,
    Grader,
    RemoteGrader,
    Trainer,
    Warning,
    ConfigurationError,
    Course,
    CourseBook,
    get_course_instance,
    batch_load,
)



from types import ModuleType
from rich._log_render import FormatTimeCallable
from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    MofNCompleteColumn,
    SpinnerColumn,
)
from rich.console import Console, ConsoleOptions, RenderResult, Group
from rich.panel import Panel
from rich.live import Live
from rich.layout import Layout
from rich.table import Table
from rich.text import Text
from rich.logging import RichHandler
from rich.highlighter import RegexHighlighter, Highlighter
from omegaconf import DictConfig
from typing import Iterable, List, Literal, Optional, Union
from contextlib import nullcontext
from dataclasses import dataclass, field
from logging import Formatter, LogRecord, StreamHandler
from logging.handlers import RotatingFileHandler
import logging
from math import ceil
from re import sub

SPINNER_STATUS = "dots"
SPINNER_TRAIN = "monkey"
SPINNER_EVAL = "moon"

REFRESH_FREQ = 12.5
PANEL_WIDTH = 100

_FILE_FORMAT = "[%(asctime)s][%(name)s][%(levelname)s] %(message)s"
_RICH_FORMAT = "%(message)s"
_STRM_FORMAT = lambda x: f"[cyan dim]%(asctime)s[/] [{x}]%(levelname)s[/] %(message)s"
_DATE_FORMAT = "%y-%m-%d %H:%M:%S"


class LogHandler(logging.Handler):
    colors = {
        10: "green",  # DEBUG
        20: "blue",  # INFO
        30: "yellow",  # WARNING
        40: "red",  # ERROR
        50: "bold red",  # CRITICAL
    }

    def __init__(self, level=0) -> None:
        super().__init__(level)

    def highlight(self, msg: str):
        return sub(
            "\bEvaluation Result\b",
            "\b[bold red]Evaluation Result[/]\b",
            sub("\bTraining State\b", "\b[bold green]Training State[/]\b", msg),
        )

    def emit(self, record: LogRecord) -> None:
        self.setFormatter(
            Formatter(_STRM_FORMAT(self.colors[record.levelno]), _DATE_FORMAT)
        )
        logview.flush(self.highlight(self.format(record)) + "\n")


class LogFileHandler(RotatingFileHandler):
    def __init__(
        self,
        filename,
        mode: str = "a",
        maxBytes: int = 0,
        backupCount: int = 0,
        encoding: str | None = None,
        delay: bool = False,
        errors: str | None = None,
    ) -> None:
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay, errors)
        self.setFormatter(Formatter(_FILE_FORMAT, _DATE_FORMAT))


class PreLogHandler(RichHandler):
    def __init__(
        self,
        level: int | str = logging.NOTSET,
        console: Console | None = None,
        *,
        show_time: bool = True,
        omit_repeated_times: bool = True,
        show_level: bool = True,
        show_path: bool = True,
        enable_link_path: bool = True,
        highlighter: Highlighter | None = None,
        markup: bool = False,
        rich_tracebacks: bool = False,
        tracebacks_width: int | None = None,
        tracebacks_extra_lines: int = 3,
        tracebacks_theme: str | None = None,
        tracebacks_word_wrap: bool = True,
        tracebacks_show_locals: bool = False,
        tracebacks_suppress: Iterable[str | ModuleType] = ...,
        locals_max_length: int = 10,
        locals_max_string: int = 80,
        log_time_format: str | FormatTimeCallable = "[%x %X]",
        keywords: List[str] | None = None,
    ) -> None:
        super().__init__(
            level,
            console,
            show_time=show_time,
            omit_repeated_times=omit_repeated_times,
            show_level=show_level,
            show_path=show_path,
            enable_link_path=enable_link_path,
            highlighter=highlighter,
            markup=markup,
            rich_tracebacks=rich_tracebacks,
            tracebacks_width=tracebacks_width,
            tracebacks_extra_lines=tracebacks_extra_lines,
            tracebacks_theme=tracebacks_theme,
            tracebacks_word_wrap=tracebacks_word_wrap,
            tracebacks_show_locals=tracebacks_show_locals,
            tracebacks_suppress=tracebacks_suppress,
            locals_max_length=locals_max_length,
            locals_max_string=locals_max_string,
            log_time_format=log_time_format,
            keywords=keywords,
        )
        self.setFormatter(Formatter(_RICH_FORMAT, _DATE_FORMAT))


@dataclass
class LogView:
    lines: list[str] = field(default_factory=list)

    def __render(self, width: int, height: int):
        rendered = []
        wlmt, hlmt = max(width - 4, 1), max(height - 5, 1)
        for line in self.lines:
            num_line = ceil(len(line) / wlmt)
            if num_line <= 1:
                rendered.append(line)
            else:
                rendered += [line[wlmt * l : wlmt * (l + 1)] for l in range(num_line)]
        return rendered[-hlmt:]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield "\n".join(self.__render(options.size.width, options.size.height))

    def flush(self, msg: str):
        self.lines += msg.strip().split("\n")
        live.refresh()

def progress_bar(spinner: str, **kwargs):
    return Progress(
        TextColumn("{task.description}"),
        SpinnerColumn(spinner),
        BarColumn(bar_width=None),
        TaskProgressColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
        **kwargs,
        refresh_per_second=REFRESH_FREQ,
        expand=True,
    )


def init_display(console: Console = None, silent: bool = False):
    if console is None:
        console = Console()

    global status, progress1, progress2, ebar, sbar, vbar, g_console, live, main_layout, logview

    g_console = console
    status = console.status(
        "Initiating...", spinner=SPINNER_STATUS, refresh_per_second=REFRESH_FREQ
    )
    progress1 = progress_bar(SPINNER_TRAIN, console=console)
    progress2 = progress_bar(SPINNER_EVAL, console=console)

    logview = LogView()

    if silent:

        live = Live(Group(status, progress1, progress2))
        live.start()

    else:
        main_layout = Layout()
        main_layout.split_column(
            Layout(name="Log"),
            Layout(name="Progress", size=5),
        )

        main_layout["Progress"].update(
            Panel(Group(status, progress1, progress2), width=PANEL_WIDTH)
        )
        main_layout["Log"].update(logview)

        live = Live(main_layout, console=g_console, refresh_per_second=REFRESH_FREQ)
        live.start()

    ebar = progress1.add_task("Epoch", visible=False)
    sbar = progress1.add_task("Update Step", visible=False)
    vbar = progress2.add_task("Inference", visible=False)

    return console

#PROGRESS

def bar(
    task: Literal["train", "eval"],
    total_epochs=None,
    total_steps=None,
    start_epoch=0,
    start_step=0,
    msg=None,
):
    global ebar_restore

    ebar_restore = total_epochs is not None

    if task == "train":
        status.update("Training...") if msg is None else status.update(msg)
        if total_epochs is not None:
            progress1.update(ebar, total=total_epochs, completed=start_epoch, visible=True)
        progress1.update(sbar, total=total_steps, completed=start_step, visible=True)
    else:
        status.update("Evaluating...") if msg is None else status.update(msg)
        progress1.update(ebar, visible=False)
        progress1.update(sbar, visible=False)
        progress2.reset(vbar, total=total_steps, completed=start_step, visible=True)

    return nullcontext()


def advance(bar_type: Literal["epoch", "step", "eval"], steps: int = 1) -> None:
    if bar_type == "epoch":
        progress1.advance(ebar, steps)
    elif bar_type == "step":
        progress1.advance(sbar, steps)
    else:
        progress2.advance(vbar, steps)


def reset(bar_type: Literal["epoch", "step", "eval"], steps: int = 0) -> None:
    if bar_type == "epoch":
        progress1.reset(ebar, completed=steps)
    elif bar_type == "step":
        progress1.reset(sbar, completed=steps)
    else:
        progress2.reset(vbar, completed=steps)


def end(task: Literal["train", "eval"], restore: bool = False):
    if task == "train":
        status.update(f"Completed!", spinner="smiley")
        progress1.stop()
        progress2.stop()
    else:
        progress2.stop()
        if restore:
            progress2.update(vbar, visible=False)
            status.update("Training...")
            if ebar_restore:
                progress1.update(ebar, visible=True)
            progress1.update(sbar, visible=True)
        else:
            progress1.stop()
            status.update("Completed!", spinner="smiley")

def initiated() -> bool:
    global g_console
    return g_console is None



"""statelogger module includes loggers to record any information about
experiments you conduct. The basic feature is to log outputs to stdout
just like builtin package 'logging', but also it can log to files or
other tools such as Tensorboard or MLFlow."""

import logging
from logging import Formatter, LogRecord, StreamHandler
from datetime import datetime
from typing import Callable, Literal, TypeVar, Union, Any
from pathlib import Path
from omegaconf import DictConfig, OmegaConf
import json
import csv
import openpyxl
from openpyxl import Workbook
from abc import abstractmethod
import os
import sys
from rich.traceback import install
from rich.console import Console
from .style import init_display, LogHandler, PreLogHandler, LogFileHandler
from datetime import datetime

install(show_locals=True)

FILE_HND_MAX_BYTES = 10_485_760
FILE_HND_BCK_COUONT = 1

T_co = TypeVar("T_co", covariant=True)

g_logger = None

class Logger:
    @abstractmethod
    def info(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def debug(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def warning(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def error(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def critical(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def log(self, level, msg, *args, **kwargs):
        pass

    @abstractmethod
    def exception(self, msg, *args, **kwargs):
        pass

    @abstractmethod
    def log_state(self, data: dict) -> None:
        pass


class DummyLogger(Logger):
    r"""DummyLogger just prints to stdout every log messages. It doesn't
    keep any training state or output any kind of results as files. If
    you need to record any information about experiments, You should
    consider using other Loggers such as StateLogger."""

    def __init__(
        self, name, level, conf, silent: bool = False, inactive: bool = False
    ) -> None:
        self.name = name
        self.level = level
        self.conf = conf
        self.silent = conf.silent
        if silent:
            self.silent = True
        self.inactive = inactive
        if inactive:
            self.silent = True

        self.init_console_logger(name, level, conf.logging_dir)

    def init_console_logger(
        self, name: str, level: int = 0, logging_dir: Union[str, Path] = None
    ) -> None:

        if isinstance(logging_dir, str):
            logging_dir = Path(logging_dir)

        self.logging_dir = logging_dir

        logging_dir.mkdir(parents=True, exist_ok=True)

        rich_handler = PreLogHandler(
            rich_tracebacks=True, show_path=False, tracebacks_suppress=[]
        )

        file_handler = LogFileHandler(
            logging_dir.joinpath("eniat.log"),
            maxBytes=FILE_HND_MAX_BYTES,
            backupCount=1,
        )

        logging.basicConfig(level=level, handlers=[rich_handler, file_handler])
        self.console = logging.getLogger(name)
        self.rich_console = None

    def load_rich_console(self) -> None:
        self.rich_console = init_display(silent=self.silent)
        file_handler = LogFileHandler(
            self.logging_dir.joinpath("eniat.log"),
            maxBytes=FILE_HND_MAX_BYTES,
            backupCount=1,
        )
        logging.basicConfig(
            level=self.level,
            handlers=[LogHandler(self.level), file_handler],
            force=True,
        )

    def be_silent(self) -> None:
        self.silent = True

    def be_inactive(self) -> None:
        self.inactive = True
        self.silent = True

    def reload(self, name=None, level=None, logging_dir=None):
        self.init_console_logger(
            name if name is not None else self.name,
            level if level is not None else self.level,
            logging_dir if logging_dir is not None else self.conf.logging_dir,
        )

    def info(self, msg, *args, **kwargs):
        return self.console.info(msg, *args, **kwargs) if not self.silent else None

    def debug(self, msg, *args, **kwargs):
        return self.console.debug(msg, *args, **kwargs) if not self.silent else None

    def warning(self, msg, *args, **kwargs):
        return self.console.warning(msg, *args, **kwargs) if not self.silent else None

    def error(self, msg, *args, **kwargs):
        return self.console.error(msg, *args, **kwargs) if not self.silent else None

    def critical(self, msg, *args, **kwargs):
        return self.console.critical(msg, *args, **kwargs) if not self.silent else None

    def log(self, level, msg, *args, **kwargs):
        return (
            self.console.log(level, msg, *args, **kwargs) if not self.silent else None
        )

    def exception(self, msg, *args, **kwargs):
        return self.console.exception(msg, *args, **kwargs) if not self.silent else None

    def log_state(self, data: dict) -> None:
        return None


class StateLogger(DummyLogger):
    """StateLogger keeps tracks of training states and can output the
    results as files. Specifically, data transferred by 'log_state'
    will be recorded into State instance."""

    class StateTable:
        def __init__(self) -> None:
            self.__state = {}
            self.__e = self.__state["epoch"] = {}
            self.__s = self.__state["step"] = {}
            self.__n = self.__state["global"] = []

        def add(self, data: dict, epoch: int = None, step: int = None) -> None:
            if epoch is None:
                if step is None:
                    self.__n.append(data)
                else:
                    if step in self.__s:
                        self.__s[step].append(data)
                    else:
                        self.__s[step] = [data]
            else:
                if not epoch in self.__e:
                    self.__e[epoch] = {"data": []}

                if step is None:
                    self.__e[epoch]["data"].append(data)
                else:
                    if step in self.__e[epoch]:
                        self.__e[epoch][step].append(data)
                    else:
                        self.__e[epoch][step] = [data]

        def get(self) -> T_co:
            pass

        def to_json(self, path: Union[Path, str]) -> None:
            with open(path, "w", encoding="utf8") as f:
                json.dump(self.__state, f, ensure_ascii=False, indent=4)

        def to_csv(self, path: Union[Path, str]) -> None:
            pass

        def to_xls(self, path: Union[Path, str]) -> None:
            pass

    def __init__(
        self, name: str, level: int = 0, conf: DictConfig = None, silent: bool = False
    ) -> None:
        super().__init__(name, level, conf, silent)
        self.__state = self.StateTable()

    def setLevel(self, level):
        return self.console.setLevel(level)

    def isEnabledFor(self, level):
        return self.console.isEnabledFor(level)

    def getEffectiveLevel(self, level):
        return self.console.getEffectiveLevel(level)

    def addFilter(self, filter):
        return self.console.addFilter(filter)

    def removeFilter(self, filter):
        return self.console.removeFilter(filter)

    def filter(self, record):
        return self.console.filter(record)

    def addHandler(self, hdlr):
        return self.console.addHandler(hdlr)

    def removeHandler(self, hdlr):
        return self.console.removeHandler(hdlr)

    def handle(self, record):
        return self.console.handle(record)

    def hasHandlers(self):
        return self.hasHandlers()

    def findCaller(self, stack_info=False, stacklevel=1):
        return self.console.findCaller(stack_info, stacklevel)

    def check_loss_policy(self, _t: int, _u: str) -> bool:
        if _t == None or _u == None or self.conf.log_interval == 0:
            return True
        return (_u == self.conf.unit) and (_t % self.conf.log_interval) == 0

    @property
    def state(self):
        return self.__state

    def log_state(
        self,
        data: dict,
        epoch: int = None,
        step: int = None,
        unit: Literal["epoch", "step"] = "epoch",
        training_state: bool = True,
        to_json: bool = None,
        to_xls: bool = None,
        to_csv: bool = None,
        silent: bool = False,
    ):

        if (
            not self.check_loss_policy(epoch if unit == "epoch" else step, unit)
            or self.inactive
        ):
            return

        self.__state.add(data, epoch, step)

        if to_json if to_json is not None else self.conf.json:
            self.__state.to_json(
                Path(self.conf.logging_dir).joinpath("training_state.json").absolute()
            )
        if to_xls if to_xls is not None else self.conf.xls:
            raise NotImplementedError("sorry. xls is not supported yet.")
        if to_csv if to_csv is not None else self.conf.csv:
            raise NotImplementedError("Sorry, csv is not supported yet.")

        if not silent:
            _str = (
                f"{'Training State' if training_state else 'Evaluation Result'} ({unit.capitalize()} {epoch if unit=='epoch' else step})\n"
                + "\n".join(
                    [
                        f"                       {item[0]} : {item[1]}"
                        for item in data.items()
                    ]
                )
            )
            self.info(_str, extra={"markup": True})


class TensorboardLogger(StateLogger):
    """TensorboardLogger literally features Tensorboard from torch.utils.
    This inherits StateLogger, thus you can record Tensorboard and files at
    the same time if you wish."""

    def __init__(
        self, name: str, conf: DictConfig = None, resume_path: str = None
    ) -> None:
        super().__init__(name, conf, resume_path)

    def prepare(self) -> None:
        tb = __import__("torch.utils.tensorboard", fromlist=["SummaryWriter"])
        self.tb = tb.SummaryWriter(
            Path(self.conf.logging_dir).joinpath("tensorboard")
        )

    def reload(self, *args, **kwargs) -> None:
        super().reload(*args, **kwargs)
        self.prepare()

    def log_state(
        self,
        data: dict,
        epoch: int = None,
        step: int = None,
        unit: Literal["epoch", "step"] = "epoch",
        training_state: bool = True,
        to_json: bool = False,
        to_xls: bool = False,
        to_csv: bool = False,
        silent: bool = False,
    ):
        if not self.check_loss_policy(epoch if unit == "epoch" else step, unit) or self.inactive:
            return
        if len(data) == 1:
            data = list(data.items())[0]
            self.log_scalar(data[0], data[1], epoch if unit == "epoch" else step, silent=silent)
        elif len(data) > 1:
            tag = "Trainig state" if training_state else "Evaluation Result"
            self.log_scalars(data, epoch if unit == "epoch" else step, tag, silent=silent)

        super().log_state(data, epoch, step, unit, training_state, to_json, to_xls, to_csv, silent=True)

    def log_scalar(self, key: str, value, step: int, silent: bool = False):
        if self.inactive:
            return
        self.tb.add_scalar(key, value, step)
        if not silent:
            self.console.info("Current state logged\n" + key + ": " + str(value))

    def log_scalars(
        self, key_value: dict, step: int, tag: str = None, silent: bool = False
    ):
        if self.inactive:
            return
        self.tb.add_scalars(tag, key_value, step)
        if not silent:
            self.console.info(
                "Current state logged\n"
                + str(f"{key} : {value}\n" for key, value in key_value.items())
            )

    def log_image(
        self,
        image,
        tag: str = None,
        step: int = None,
        format: str = None,
        filepath: str = None,
        silent: bool = False,
    ):
        if self.inactive:
            return
        self.tb.add_image(tag, image, step, format)
        if not silent:
            self.console.info(f"Logged an image to tensorboard")

    def log_figure(
        self,
        figure,
        tag: str = None,
        step: int = None,
        close: bool = None,
        filepath: str = None,
        silent: bool = False,
    ):
        if self.inactive:
            return
        self.tb.add_figure(tag, figure, step, close)
        if not silent:
            self.console.info(f"Logged a figure to tensorboard")


class MLFlowLogger(StateLogger):
    """TensorboardLogger literally features Tensorboard from torch.utils.
    This inherits StateLogger, thus you can record Tensorboard and files at
    the same time if you wish."""

    def __init__(
        self,
        name: str,
        conf: DictConfig = None,
        resume_path=None,
    ) -> None:
        super().__init__(name, conf, resume_path)

    def prepare(
        self,
    ) -> None:

        self.uri = self.conf.mlflow.uri
        self.mf = __import__("mlflow")
        self.mf.set_tracking_uri(self.uri)

        if self.conf.mlflow.autolog:
            # Refer to autolog reference and fix here
            self.mf.autolog()
        elif self.conf.mlflow.run_id is not None:
            self.current_run = self.mf.start_run(
                self.conf.mlflow.run_id, ru_name=self.conf.mlflow.run_name
            )
        else:
            if self.conf.mlflow.exp_id is not None:
                self.experiment = self.mf.set_experiment(
                    experiment_id=self.conf.mlflow.exp_id
                )
            else:
                self.experiment = self.mf.set_experiment(
                    experiment_name=self.conf.mlflow.exp_name
                )
            self.current_run = self.mf.start_run(experiment_id=self.conf.mlflow.exp_id)

    def reload(self, *args, **kwargs) -> None:
        super().reload(*args, **kwargs)
        self.prepare()

    def __del__(self) -> None:

        if not "mf" in self.__dict__:
            return

        if self.mf is not None and self.mf.active_run() is not None:
            self.mf.end_run()

    def log_scalar(self, key: str, value, step: int, silent: bool = False):
        if self.inactive:
            return
        self.mf.log_metric(key, value, step)
        if not silent:
            self.console.info("Current state logged\n" + key + ": " + str(value))

    def log_scalars(
        self, key_value: dict, step: int, tag: str = None, silent: bool = False
    ):
        if self.inactive:
            return
        self.mf.log_metrics(key_value, step)
        if not silent:
            self.console.info(
                "Current state logged\n"
                + (f"{key} : {value}\n" for key, value in key_value.items())
            )

    def log_image(
        self,
        image,
        tag: str = None,
        step: int = None,
        format: str = None,
        filepath: str = None,
        silent: bool = False,
    ):
        if self.inactive:
            return
        self.mf.log_image(image, filepath)
        if not silent:
            self.console.info(f"Logged an image to {self.mf.get_tracking_uri()}")

    def log_figure(
        self,
        figure,
        tag: str = None,
        step: int = None,
        close: bool = None,
        filepath: str = None,
        silent: bool = False,
    ):
        if self.inactive:
            return
        self.mf.log_figure(figure, filepath)
        if not silent:
            self.console.info(f"Logged a figure to {self.mf.get_tracking_uri()}")

    def log_state(
        self,
        data: dict,
        epoch: int = None,
        step: int = None,
        unit: Literal["epoch", "step"] = "epoch",
        training_state: bool = True,
        to_json: bool = False,
        to_xls: bool = False,
        to_csv: bool = False,
        silent: bool = False,
    ):

        if not self.check_loss_policy(data, unit) or self.inactive:
            return
        if len(data) == 1:
            data = list(data.items())[0]
            self.log_scalar(data[0], data[1], epoch if unit == "epoch" else step, silent=silent)
        elif len(data) > 1:
            self.log_scalars(data, epoch if unit == "epoch" else step, "Training State" if training_state else "Evaluation Result", silent=silent)

        super().log_state(data, epoch, step, unit, training_state, to_json, to_xls, to_csv, True)


class TotalLogger(StateLogger):

    console = None
    mf_logger = None
    tb_logger = None

    def __init__(self, *args, **kwargs) -> None:
        self.init_logger(*args, **kwargs)

    def __del__(self) -> None:
        if not self.autolog:
            self.mf_logger.end_run()
        self.tb_logger.close()

    def get_mlflow_logger(self):
        return self.mf_logger

    def get_tensorboard_logger(self):
        return self.tb_logger

    def init_logger(
        self,
        name: str,
        file: bool = True,
        mlflow: bool = True,
        remote_model_save: bool = False,
        tensorboard: bool = True,
        autolog: bool = False,
        autolog_options: dict = False,
        uri: str = "",
        experiment_id: str = None,
        run_id: str = None,
        experiment_name: str = None,
    ):

        self.console = logging.getLogger(name)

        self.experiment_id = experiment_id
        self.run_id = run_id
        self.file = file
        self.mlflow = mlflow
        self.remote_model = remote_model_save
        self.tensorboard = tensorboard
        self.autolog = autolog
        self.uri = uri

        if mlflow:
            self.mf_logger = __import__("mlflow")
            self.mf_logger.set_tracking_uri(uri)
            if autolog:
                self.mf_logger.autolog(**autolog_options)
            else:

                if experiment_id:
                    self.mf_logger.set_experiment(experiment_id=experiment_id)
                else:
                    self.mf_logger.set_experiment(experiment_name=experiment_name)

                self.mf_logger.start_run(run_id)

        if tensorboard:
            self.tb = __import__("torch.utils.tensorboard", fromlist=["torch.utils"])
            self.tb_logger = self.tb.SummaryWriter("tensorboard")

    def log_scalar(self, key: str, value, step: int):
        if self.inactive:
            return
        if self.mlflow and not self.autolog:
            self.mf_logger.log_metric(key, value, step)
        if self.tensorboard:
            self.tb_logger.add_scalar(key, value, step)
        self.console.info(key + ": " + value)

    def log_scalars(self, key_value: dict, step: int, tag: str = None):
        if self.inactive:
            return
        if self.mlflow and not self.autolog:
            self.mf_logger.log_metrics(key_value, step)
        if self.tensorboard:
            self.tb_logger.add_scalars(tag, key_value, step)
        for k, v in key_value.items():
            self.console.info(k + ": " + v)

    def log_image(
        self,
        image,
        tag: str = None,
        step: int = None,
        format: str = None,
        filepath: str = None,
    ):
        if self.inactive:
            return
        if self.mlflow and not self.autolog:
            self.mf_logger.log_image(image, filepath)
            self.console.info(f"Logged an image to {self.mf_logger.get_tracking_uri()}")
        if self.tensorboard:
            self.tb_logger.add_image(tag, image, step, format)
            self.console.info(f"Logged an image to tensorboard")

    def log_figure(
        self,
        figure,
        tag: str = None,
        step: int = None,
        close: bool = None,
        filepath: str = None,
    ):
        if self.inactive:
            return
        if self.mlflow and not self.autolog:
            self.mf_logger.log_figure(figure, filepath)
            self.console.info(f"Logged a figure to {self.mf_logger.get_tracking_uri()}")
        if self.tensorboard:
            self.tb_logger.add_figure(tag, figure, step, close)
            self.console.info(f"Logged a figure to tensorboard")


def load_logger(logger_conf: DictConfig) -> Logger:
    logger = getattr(sys.modules[__name__], logger_conf.type)(
        logger_conf.name, logger_conf.level, logger_conf
    )
    return logger

def init_logger(
        log_cls = StateLogger,
        conf:DictConfig = None,
        **kwargs
    ) -> Logger:
    global g_logger
    if g_logger is not None:
        if isinstance(g_logger, log_cls):
            return g_logger
        else:
            g_logger = None
    if conf is not None:
        g_logger = load_logger(conf)
    else:
        g_logger = log_cls(**kwargs)
    return g_logger
    
def get_logger() -> Logger:
    return g_logger or init_logger()


from typing import TypeVar, Literal, Union
from .grader import TorchGrader
from .predictor import TorchPredictor, to_tensor, distributed
from ..core import Trainer, CourseBook
from .learner import TorchLearner
from ..utils import StateLogger, bar, advance, reset, end, init_logger
from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    MofNCompleteColumn,
    SpinnerColumn,
)
from rich.live import Live
import torch
import torch.distributed as dist
from torch.cuda.amp import GradScaler
import os
from omegaconf import DictConfig
from pathlib import Path
import warnings
from contextlib import nullcontext
import itertools
import numpy as np

T = TypeVar("T", bound=TorchLearner)
C = TypeVar("C", bound=CourseBook)
L = TypeVar("L", bound=StateLogger)
G = TypeVar("G", bound=TorchGrader)

class TorchTrainer(TorchPredictor, Trainer):
    r"""PyTorch compatible trainer class.
    Automatically manage training steps, logging, and saving checkpoints.
    It only occupies one device(GPU)"""

    def __init__(
        self,
        conf: DictConfig = None,
        log=None,
        grader=None,
        course=None,
        learner=None,
        loss_fn=None,
        optimizer=None,
        scheduler=None,
    ) -> None:

        super(TorchTrainer, self).__init__(conf.env)
        self.conf = conf if not "trainer" in conf else conf.trainer
        if log is not None:
            self.log = log
        elif "logger" in conf:
            self.log = init_logger()
        self.grader = grader
        self.course = course
        self.learner = learner
        self.loss = loss_fn
        self.opt = optimizer
        self.sch = scheduler

        warnings.showwarning = lambda *args: self.log.warning(args[0])

    def _save_model(self, timestep: int = None) -> None:
        Path(os.path.join(self.conf.output_dir, "checkpoints")).mkdir(
            parents=True, exist_ok=True
        )
        torch.save(
            self.learner.model.state_dict(),
            os.path.join(
                self.conf.output_dir,
                f"checkpoints/model_{timestep}.cpt"
                if timestep
                else "checkpoints/model.cpt",
            ),
        )

    def _save_state(self, timestep: int, filename: str = None) -> None:
        train_state = {}
        train_state["scheduler"] = (
            self.sch.state_dict() if self.sch is not None else None
        )
        train_state["optimizer"] = self.opt.state_dict()
        train_state["unit"] = self.conf.scheme.unit
        train_state["rng_state"] = self.get_rand_state()
        train_state["timestep"] = timestep
        train_state["batch_size"] = self.conf.loader.batch_size
        train_state["total_iters"] = self.conf.scheme.total_iters
        train_state["env"] = self.conf.env
        Path(os.path.join(self.conf.output_dir, "checkpoints")).mkdir(
            parents=True, exist_ok=True
        )
        torch.save(
            train_state,
            os.path.join(
                self.conf.output_dir,
                f"checkpoints/state_{timestep}.cpt" if filename is None else filename,
            ),
        )

    def _save_checkpoint(
        self, timestep: int, unit: Literal["epoch", "step"], force: bool = False
    ) -> None:
        if not force and (
            self.conf.scheme.unit != unit or (timestep % self.conf.scheme.save_interval)
            if self.conf.scheme.save_interval
            else True
        ):
            return
        self._save_model(timestep)
        self._save_state(timestep)
        self.log.info(f"Checkpoint at {unit} {timestep} saved.")

    def __chk(self, epoch: int = None, step: int = None):
        if self.conf.scheme.unit == "step" and step is not None:
            if step % self.conf.scheme.save_interval == 0:
                if dist.is_initialized():
                    self.opt.consolidate_state_dict(0)
                if not dist.is_initialized() or dist.get_rank() == 0:
                    self._save_checkpoint(step, "step")
        elif self.conf.scheme.unit == "epoch" and epoch is not None:
            if epoch % self.conf.scheme.save_interval == 0:
                if dist.is_initialized():
                    self.opt.consolidate_state_dict(0)
                if not dist.is_initialized() or dist.get_rank() == 0:
                    self._save_checkpoint(epoch, "epoch")
                    
    @distributed
    def fit(
        self,
        device: int = 0,
        global_rank: int = 0,
        silent: bool = False,
        final: bool = True,
    ):

        resumed = self.prepare(
            device,
            log=self.log,
        )

        total_step = 0
        saved = False

        loss = None
        s_loss = torch.zeros(2, device=device, requires_grad=False)
        e_loss = torch.zeros(2, device=device, requires_grad=False)

        if resumed:

            if self.conf.scheme.unit == "step":
                total_step = self.conf.scheme.init_iter

            self.log.info(f"Training state ({self.conf.scheme.init_iter} \
                          {self.conf.scheme.unit}) resumed.")

        with bar(
            "train",
            total_epochs=self.conf.scheme.total_iters
            if self.conf.scheme.unit == "epoch"
            else None,
            total_steps=self.conf.scheme.total_iters
            if self.conf.scheme.unit == "step"
            else len(self.loader),
            start_epoch=self.conf.scheme.init_iter
            if self.conf.scheme.unit == "epoch"
            else None,
            start_step=self.conf.scheme.init_iter
            if self.conf.scheme.unit == "step"
            else None,
        ) if (not dist.is_initialized() or dist.get_rank() == 0) else nullcontext():

            for epoch in range(
                self.conf.scheme.init_iter + 1,
                self.conf.scheme.total_iters + 1) if self.conf.scheme.unit == "epoch" else itertools.count(1):

                if self.conf.scheme.unit == "epoch":
                    if not dist.is_initialized() or dist.get_rank() == 0:
                        reset("step")

                for current_step, batch in enumerate(self.loader):

                    total_step += 1

                    # Forward
                    batch = to_tensor(
                        batch, dtype=self.conf.scheme.dtype, device=device
                    )
                    self.learner.model.train(True)
                    self.opt.zero_grad()

                    if self.conf.scheme.precision is None:
                        pred = self.learner.fit(batch[0], device, self.log)
                        cur_loss = self.loss(pred, batch[1].to(device))
                    else:
                        with torch.autocast(
                            torch.device(device),
                            dtype=getattr(torch, self.conf.scheme.precision),
                        ):
                            pred = self.learner.fit(batch[0], device, self.log)
                            cur_loss = self.loss(pred, batch[1].to(device))

                    _back = False

                    # Backward
                    if self.conf.scheme.update_interval:
                        if loss is None:
                            loss = cur_loss
                        else:
                            loss += cur_loss

                        if (current_step + 1) % self.conf.scheme.update_interval:
                            _back = True
                    else:
                        loss = cur_loss
                        _back = True

                    if _back or current_step == len(self.loader) - 1:

                        s_loss[1] = batch[0].size(dim=0)
                        s_loss[0] = loss.item() * s_loss[1]

                        e_loss[0] += s_loss[0]
                        e_loss[1] += s_loss[1]

                        if self.conf.scheme.gradient_scale:
                            scaler = GradScaler(loss)
                            scaler.scale(loss).backward()
                            scaler.step(self.opt)
                            scaler.update()
                        else:
                            scaler = None
                            loss.backward()
                            self.opt.step()

                        loss = None

                    self.learner.model.train(False)

                    if dist.is_initialized():
                        dist.reduce(s_loss, 0, dist.ReduceOp.SUM)

                    if not dist.is_initialized() or dist.get_rank() == 0:
                        self.log.log_state(
                            {
                                "Training Loss": (s_loss[0] / s_loss[1]).item(),
                                "Learning Rate": np.array(self.sch.get_last_lr())
                                if self.sch
                                else 0,
                            },
                            epoch,
                            total_step,
                            "step",
                        )

                    if not dist.is_initialized() or dist.get_rank() == 0:
                        advance("step")

                    # Evaluation (step)
                    if self.grader is not None:
                        _rng_state = self.get_rand_state()
                        if self.grader.eval(
                            device,
                            global_rank,
                            silent,
                            False,
                            learner=self.learner,
                            timestep=total_step,
                            unit="step",
                            env_conf=self.conf.env,
                        ):
                            self.set_rand_state(_rng_state)

                    # Checkpoint (step)
                    self.__chk(step=total_step)

                    if self.conf.scheme.unit == "step":
                        if self.sch:
                            self.sch.step()
                        if self.conf.scheme.total_iters <= total_step:
                            break

                # STEP OVER ==========================================================

                if self.conf.scheme.unit == "step" and \
                    self.conf.scheme.total_iters <= total_step:
                    break

                # Scheduler
                if self.sch and self.conf.scheme.unit == "epoch":
                    self.sch.step()

                if dist.is_initialized():
                    dist.reduce(e_loss, 0, dist.ReduceOp.SUM)
                
                if not dist.is_initialized() or dist.get_rank() == 0:
                    self.log.log_state(
                        {
                            "Training Loss": (e_loss[0] / e_loss[1]).item(),
                            "Learning Rate": np.array(self.sch.get_last_lr()) if self.sch else 0,
                        },
                        epoch,
                        total_step,
                        "epoch",
                    )

                e_loss[0] = e_loss[1] = 0

                if not dist.is_initialized() or dist.get_rank() == 0:
                    advance("epoch")

                self.__chk(epoch=epoch)

                # Evaluation (epoch)
                if self.grader is not None:
                    _rng_state = self.get_rand_state()
                    if self.grader.eval(
                        device,
                        global_rank,
                        silent,
                        False,
                        learner=self.learner,
                        timestep=epoch,
                        unit="epoch",
                        env_conf=self.conf.env,
                    ):
                        self.set_rand_state(_rng_state)

        # loop over
        if not dist.is_initialized() or dist.get_rank() == 0:
            end("train")

        if dist.is_initialized():
            self.opt.consolidate_state_dict()

        if not dist.is_initialized() or dist.get_rank() == 0:
            self._save_checkpoint(self.conf.scheme.total_iters, self.conf.scheme.unit, True)

        if dist.is_initialized():
            dist.destroy_process_group()



from .statelogger import (
    Logger,
    StateLogger,
    TensorboardLogger,
    MLFlowLogger,
    TotalLogger,
    load_logger,
    init_logger,
    get_logger,
)
from .cli import parser, init
from .conf import (
    load_conf,
    recursive_merge,
    recursive_merge_by_path,
    import_by_file,
    instantiate,
    load_class,
)
from .manager import Manager
from .style import *



from typing import TypeVar
from abc import abstractmethod
from ..core import Learner
from sklearn.base import BaseEstimator


B = TypeVar("B", bound=BaseEstimator)


class ScikitLearner(Learner):
    def __init__(self, model: B) -> None:
        super(ScikitLearner).__init__()

        self.model = model

    @abstractmethod
    def predict(self, X):
        pass

    @abstractmethod
    def fit_step(self):
        pass

    def get_model(self) -> B:
        return self.model



from typing import TypeVar, Generic, Union, Any
from abc import abstractmethod
from ..core import Learner
import torch
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from ..utils import instantiate, Logger, load_class
from omegaconf import DictConfig
from pathlib import Path

T_co = TypeVar("T_co", covariant=True)
O = TypeVar("O", bound=Optimizer)
L = TypeVar("L", bound=Logger)


def load_learner(learner_conf: DictConfig):
    return load_class(learner_conf.path, learner_conf.cls)(learner_conf)

class TorchLearner(Learner, Generic[T_co]):
    def __init__(
        self,
        conf: DictConfig = None,
        model: Module = None,
    ) -> None:
        super(TorchLearner).__init__()
        self.conf = conf
        self.model = model

    @abstractmethod
    def fit(self, batch: Tensor, device: int, logger):
        pass

    @abstractmethod
    def predict(self, batch: Tensor, device: int, logger):
        pass

    def prepare(self, log: L = None) -> None:
        self.model = instantiate(
            self.conf.model
        )

    def resume_model(self, state=None, path: Union[str, Path] = None) -> None:
        if state:
            self.model.load_state_dict(state)
        else:
            with open(path, "rb") as f:
                self.model.load_state_dict(torch.load(f))

    def resume(self) -> bool:
        if self.conf.resume_path:
            if isinstance(self.conf.resume_path, str):
                resume_path = Path(self.conf.resume_path)
            if resume_path.is_file():
                self.resume_model(path=resume_path)
            else:
                raise FileNotFoundError(
                    f"Cannot open the checkpoint: {self.conf.resume_path}"
                )
        elif self.conf.resume_dir and self.conf.resume_step:
            if isinstance(self.conf.resume_dir, str):
                resume_dir = Path(self.conf.resume_dir)
            if resume_dir.is_dir() and isinstance(self.conf.resume_step, int):
                self.resume_model(
                    path=f"{resume_dir.joinpath(f'model_{self.conf.resume_step}.cpt')}"
                )
        else:
            return False
        return True

    @property
    def get_model(self):
        return self.model

    def to(self, device: Union[int, str]):
        self.model.to(device)

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def __call__(self, *args, **kwargs) -> Any:
        return self.model(*args, **kwargs)


class SupervisedLearner(TorchLearner):
    def __init__(
        self, conf:DictConfig, model: Module = None,
    ) -> None:
        super().__init__(conf, model)

    def fit(self, batch: Tensor, device: int, logger):
        x = batch
        x = x.to(device)
        model = self.model.to(device)
        return model(x)

    def predict(self, batch: Tensor, device: int, logger):
        batch = batch.to(device)
        self.model.to(device)
        return self.model(batch)



from .learner import TorchLearner, SupervisedLearner, load_learner
from .trainer import TorchTrainer, distributed
from .grader import TorchGrader, load_grader
from .predictor import TorchPredictor, to_tensor
from .scheduler import *


from typing import Literal, Union, Callable, TypeVar, Sequence
import numpy as np
import pickle as pk
from omegaconf import DictConfig
from ..utils.conf import instantiate, load_class

T_co = TypeVar("T_co", covariant=True)


class ConfigError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def get_course_instance(cfg: DictConfig, log=None, as_list: bool = False):
    if as_list:
        _courses = []
    else:
        _courses = CourseBook()

    for label in cfg:
        if "cls" in cfg[label] and cfg[label]["cls"]:
            _courses.append(
                Course(
                    label,
                    instantiate(cfg[label]),
                    transform=instantiate(cfg[label]["transform"], _partial=True),
                    cache=cfg[label]["cache"],
                )
            )
            if log:
                log.info(f"'{label}' data is loaded.")
        elif "path" in cfg[label] and cfg[label]["path"]:
            _courses.append(
                Course(
                    label,
                    data=batch_load(cfg[label]["path"], cfg[label].type),
                    transform=instantiate(cfg[label]["transform"], _partial=True),
                    cache=cfg[label["cache"]],
                )
            )
            if log:
                log.info(f"'{label}' data is loaded.")
        else:
            if log:
                log.warning(
                    f"Data(:{label}) is not loaded because the path of data is not specified."
                )
            return
    if not len(_courses):
        log.error("No data is given! Terminating the job...")
        return
    return _courses


def batch_load(paths: Union[str, Sequence[str]], type: Literal["csv", "npy", "pkl"]):
    def np_load(path: str, type: Literal["csv", "npy", "pkl"]):
        if type == "csv":
            return np.loadtxt(path, delimiter=",")
        if type == "npy":
            return np.load(path)
        if type == "pkl":
            return np.load(path, allow_pickle=True)

    if isinstance(paths, str):
        return np_load(paths, type)
    else:
        ret = None
        for path in paths:
            stack = np_load(path, type)
            if ret is None:
                ret = np.empty((0, *stack.shape[1:]))
            ret = np.concatenate((ret, stack))
        return ret


class Course:
    def __init__(
        self,
        label: str,
        data: T_co,
        target: T_co = None,
        transform: Callable = None,
        cache: bool = False,
    ) -> None:
        self.label = label
        self.data = data
        self.target = target
        self.transform = transform
        self.cache = cache

        if target and (t := len(target)) != (d := len(data)):
            raise ValueError(f"The data has different size({d}) with the target({t}).")

    def has_target(self) -> bool:
        return bool(self.target)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> T_co:
        if self.transform:
            x = self.transform(self.data[idx])
        else:
            x = self.data[idx]

        if self.target:
            return x, self.target[idx]
        else:
            return x

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Course):
            return self.label == __value.label
        else:
            raise TypeError(
                "class <Course> does not support comaprison between other types than 'Course'."
            )

    def __iter__(self):
        return iter(zip(self.data, self.target)) if self.target else iter(self.data)

    def __repr__(self) -> str:
        return (
            f"{self.label} ({self.__len__()}/{len(self.target) if self.target else 0})"
        )


class CourseBook:
    def __init__(
        self,
        courses: Union[Course, Sequence[Course]] = None,
        transforms: Union[Callable, Sequence[Callable], dict] = None,
        preprocess: bool = False,
        conf: DictConfig = None,
    ) -> None:

        self.conf = conf

        if courses:
            if isinstance(courses, Course):
                self.__courses = {courses.label: courses}
            else:
                self.__courses = {course.label: course for course in courses}
        else:
            self.__courses = {}

        if transforms:
            if callable(transforms):
                for cr in self.__courses.values():
                    if preprocess:
                        cr.data = transforms(cr.data)
                    else:
                        cr.transform = transforms
            elif isinstance(transforms, dict):
                for key, tr in transforms.items():
                    if preprocess:
                        self.__courses[key].data = tr(self.__courses[key])
                    else:
                        self.__courses[key].transform = tr
            else:  # seq(callable)
                if len(transforms) == len(courses):
                    for i, cr in enumerate(self.__courses.values()):
                        if preprocess:
                            cr.data = transforms(cr.data)
                        else:
                            cr.transform = transforms[i]
                else:
                    raise ValueError(
                        "Sequences of Courses and Transforms must have the same length"
                    )

    def load(self) -> None:
        course_list = get_course_instance(self.conf, as_list=True)
        for course in course_list:
            self.__courses[course.label] = course

    def get(self, label: str):
        return self.__courses[label]

    def append(self, course: Course = None, label: str = None, data=None):
        if course:
            self.__courses[course.label] = course
        elif data and label:
            self.__courses[label] = data
        else:
            raise ValueError(
                "At least one of 'course' or a pair of 'label' and 'data' must be given"
            )
        return self

    def __getitem__(self, key) -> Course:
        return self.__courses[key]

    def __len__(self) -> int:
        return len(self.__courses)

    def __add__(self, op):
        if isinstance(op, Course):
            return CourseBook(list(set([v for k, v in self.__courses.items()] + op)))
        elif isinstance(op, CourseBook):
            return CourseBook(
                list(
                    set(
                        [v for k, v in self.__courses.items()]
                        + [v for k, v in op.courses.items()]
                    )
                )
            )
        else:
            raise TypeError(
                "class <FullCourse> does not support addtion with types other than 'Course' or 'FullCourse'."
            )

    def __repr__(self) -> str:
        return "\n".join(
            [
                f"{k} : {{ length: {len(course)}, target: {'exists' if course.has_target() else 'none'} }}"
                for k, course in self.__courses.items()
            ]
        )



import traceback

from .core import CourseBook
from .utils import init, load_logger


def console():
    conf = init()

    if conf is None:
        return

    logger = load_logger(conf.logger)
    course = CourseBook(conf=conf.data)
    logger.debug(f"Task '{conf.task}' requested using package {conf.package}")
    try:
        if conf.package == "sklearn":
            logger.critical("scikit-learn is not yet supported.")
        elif conf.package == "torch":
            from .torch import load_learner, TorchTrainer, TorchGrader

            learner = load_learner(conf.learner)

            if conf.task == "fit" or conf.task == "train":

                grader = TorchGrader(conf.grader, logger=logger, course=course, learner=learner,)
                trainer = TorchTrainer(conf.trainer, logger, grader, course, learner)
                trainer.fit()

            if conf.task == "eval" or conf.task == "test":

                grader = TorchGrader(conf.grader, logger=logger, course=course, learner=learner,)
                grader.eval()

            if conf.task == "infer" or conf.task == "predict":
                grader = TorchGrader(conf.grader, logger=logger, course=course, learner=learner,)
                grader.predict()

        else:
            raise ValueError(f"Not a valid option (package: {conf.package})")
    except:
        logger.exception(traceback.format_exc())



# from typing import
import argparse
from datetime import datetime
from importlib.metadata import version
from pathlib import Path

from omegaconf import DictConfig, OmegaConf

from .conf import load_conf, recursive_merge, recursive_merge_by_path


def datetime_to_str() -> str:
    return datetime.now().strftime("%y%m%d_%H:%M:%S")


MLTASK = ("fit", "train", "eval", "test", "infer", "predict")

parser = argparse.ArgumentParser(
    "ENIAT",
    epilog="",
    description="ENIAT supports and boosts your machine learning experiments!",
    exit_on_error=False,
)

parser.add_argument("-v", "--version", action="store_true")

parser.add_argument(
    "-p",
    "--package",
    type=str,
    choices=["torch", "sklearn"],
    help="Based package",
)
parser.add_argument("-t", "--task", choices=MLTASK)
parser.add_argument(
    "-c", "--config", nargs="*", type=Path, help="yaml config files to parse."
)
parser.add_argument(
    "-s",
    "--silent",
    action="store_true",
    help="if True, no console output. (not affect File output)",
)
parser.add_argument(
    "-o",
    "--output_dir",
    type=Path,
    default=Path("./checkpoints"),
    help="Output directory",
)
parser.add_argument(
    "-l",
    "--logging_dir",
    type=Path,
    help="Logging directory",
)

def init() -> tuple:
    main_args, cli_args = parser.parse_known_args()

    if main_args.version:
        print("eniat", version("eniat"))
        return

    default_conf = load_conf()
    user_conf = None
    if main_args.config:
        user_conf = recursive_merge_by_path(*main_args.config)
    cli_conf = OmegaConf.from_dotlist(cli_args)

    conf = recursive_merge(default_conf, user_conf, cli_conf or OmegaConf.create({}))

    (
        conf.package,
        conf.task,
        conf.output_dir,
        conf.logger.silent,
        conf.logger.logging_dir,
    ) = (
        main_args.package or conf.package,
        main_args.task or conf.task,
        main_args.output_dir,
        main_args.silent,
        main_args.logging_dir or main_args.output_dir,
    )

    OmegaConf.resolve(conf)

    return conf



import contextlib
import functools
import os
import random
from contextlib import nullcontext
from pathlib import Path
from typing import Callable, Sequence, TypeVar, Union

import numpy as np
import torch
import torch.distributed as dist
import torch.distributed.algorithms.model_averaging.averagers as averagers
from omegaconf import DictConfig
from torch.distributed.optim import (
    DistributedOptimizer,
    PostLocalSGDOptimizer,
    ZeroRedundancyOptimizer,
)
from torch.multiprocessing import spawn
from torch.nn import DataParallel as DP
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader, DistributedSampler

from ..utils import Logger, advance, bar, end, instantiate

L = TypeVar("L", bound=Logger)

ENV_NON_DIST_TYPES = ("single", "DP")
ENV_DIST_TYPES = ("DDP", "torchrun", "FSDP")


class Hooker:
    def __init__(
        self,
        labels: Sequence[str] = [
            "before_epoch",
            "before_step",
            "after_step",
            "after_epoch",
        ],
        hooks: dict = None,
    ) -> None:

        self.hooks = {label: [] for label in labels}
        if hooks is not None:
            self.update(hooks)

    def hook(self, label: str, func: Callable) -> None:
        self.hooks[label].append(func)

    def pull(self, label: str, *args, **kwargs) -> None:
        for hook in self.hooks:
            hook(*args, **kwargs)

    def update(self, hooks: dict) -> None:
        for hook, func in hooks.items():
            if hook in self.hooks and isinstance(func, list):
                self.hooks[hook] += func

    def add_label(self, label: str) -> None:
        self.hooks[label] = []


def to_tensor(batch, dtype: str = None, device=None):
    if isinstance(batch, list):
        for data in batch:
            data = torch.as_tensor(
                data,
                dtype=getattr(torch, dtype) if dtype else torch.float32,
                device=device,
            )
    else:
        batch = torch.as_tensor(
            batch,
            dtype=getattr(torch, dtype) if dtype else torch.float32,
            device=device,
        )
    return batch


def distributed(fn: Callable) -> Callable:
    def wrapper(
        self,
        device=None,
        global_rank=None,
        silent=False,
        final=True,
        **kwargs,
    ):

        if not dist.is_initialized():

            if self.conf.env.type == "single":  # default
                device = (
                    device
                    if (device is not None)
                    else (
                        self.conf.env.device_id
                        if (isinstance(self.conf.env.device_id, int))
                        else self.conf.env.device_id[0]
                    )
                )
                return fn(self, device, device, silent, final, **(kwargs or {}))
            elif self.conf.env.type == "DP":  # DP (Data Parallel)
                return fn(self, "cuda", 0, silent, final, **(kwargs or {}))

            if self.conf.env.debug:
                os.environ["TORCH_CPP_LOG_LEVEL"] = "INFO"
                os.environ["TORCH_DISTRIBUTED_DEBUG"] = "DETAIL"

            if dist.is_torchelastic_launched():
                # torchrun
                self._torchrun_init(self)
                return fn(
                    int(os.environ["LOCAL_RANK"]),
                    int(os.environ["RANK"]),
                    silent,
                    final,
                    **(kwargs or {}),
                )
            else:
                # DDP or FSDP
                spawn(
                    self._ddp_init,
                    (
                        fn.__name__,
                        silent,
                        final,
                        *[kwargs[key] for key in kwargs],
                    ),
                    nprocs=self.conf.env.dist.local_size,
                    join=True,
                )
        else:
            if dist.get_rank() == 0:
                return fn(self, device, global_rank, silent, final, **(kwargs or {}))
            else:
                with contextlib.redirect_stdout(None):
                    with contextlib.redirect_stderr(None):
                        return fn(
                            self,
                            device,
                            global_rank,
                            True,
                            final,
                            **(kwargs or {}),
                        )

    return wrapper


class TorchPredictor:
    def __init__(self, env_configuration: DictConfig, hooks: dict = None) -> None:
        self.hooker = Hooker()

    def __rank0(self):
        return dist.is_initialized() and dist.get_rank() == 0

    def get_loader(self, data_label: str = None, dataset=None) -> DataLoader:
        if dataset is None:
            if isinstance(data_label, str):
                dataset = self.course.get(data_label)
            else:
                raise ValueError(f"Invalid data label: {data_label}")
        if self.conf.env.type in ENV_DIST_TYPES:
            return (
                DataLoader(
                    dataset,
                    batch_size=self.conf.loader.batch_size,
                    shuffle=self.conf.loader.shuffle,
                    num_workers=self.conf.loader.num_workers,
                    pin_memory=self.conf.loader.pin_memory,
                )
                if not dist.is_initialized()
                else DataLoader(
                    dataset,
                    batch_size=self.conf.loader.batch_size,
                    shuffle=self.conf.loader.shuffle,
                    num_workers=self.conf.loader.num_workers,
                    pin_memory=self.conf.loader.pin_memory,
                    sampler=DistributedSampler(
                        dataset,
                        num_replicas=self.conf.env.dist.world_size,
                        rank=self.conf.env.dist.global_rank,
                    ),
                )
            )
        else:
            return DataLoader(
                dataset,
                batch_size=self.conf.loader.batch_size,
                shuffle=self.conf.loader.shuffle,
                num_workers=self.conf.loader.num_workers,
                pin_memory=self.conf.loader.pin_memory,
            )

    def _ddp_init(
        self,
        local_rank: int,
        fname: str,
        silent: bool = False,
        final=True,
        *args,
    ) -> None:
        os.environ["LOCAL_RANK"] = str(local_rank)
        os.environ["MASTER_ADDR"] = self.conf.env.dist.master_address
        os.environ["MASTER_PORT"] = str(self.conf.env.dist.master_port)
        rank = self.conf.env.dist.global_rank + local_rank
        dev_id = 0
        if self.conf.env.device_id is not None:
            if isinstance(self.conf.env.device_id, list):
                dev_id = self.conf.env.device_id[local_rank]
            else:
                dev_id = local_rank + self.conf.env.device_id
        else:
            dev_id = local_rank
        dist.init_process_group(
            backend=self.conf.env.dist.backend,
            init_method=self.conf.env.dist.init_method,
            world_size=self.conf.env.dist.world_size,
            rank=rank,
        )
        if local_rank == 0:
            self.log.info("configured DDP environment...")
        return getattr(self, fname)(dev_id, rank, silent, final, *args)

    def _torchrun_init(self):
        self.log.info("setting torchrun environment...")
        self.conf.env.dist.local_size = int(os.environ["LOCAL_WORLD_SIZE"])
        self.conf.env.dist.local_rank = int(os.environ["LOCAL_RANK"])
        self.conf.env.dist.global_rank = int(os.environ["RANK"])
        self.conf.env.dist.world_size = int(os.environ["WORLD_SIZE"])
        dist.init_process_group(backend=self.conf.env.dist.backend)

    def get_dist_opt(self, method: str, opt=None, params=None, _lr=None, **kwargs):
        if method == "zero":
            if not params:
                raise ValueError(
                    "model parameters require for zero redundancy optimizer."
                )
            return ZeroRedundancyOptimizer(
                params,
                optimizer_class=getattr(torch.optim, type(opt).__name__),
                lr=_lr,
                **kwargs,
            )
        if method == "postlocal":
            return PostLocalSGDOptimizer(opt, averagers.PeriodicModelAverager(**kwargs))
        if method == "default":
            return DistributedOptimizer(opt, params, lr=_lr, **kwargs)
        if method is None or method == "none":
            return opt
        raise ValueError(
            f'Distributed optimizer "{method}" is not valid.'
            "Configure as one of following choices:[zero, postlocal]"
        )

    def rand_all(self, seed) -> None:
        if self.conf.env.type != "single":
            torch.cuda.manual_seed_all(seed)
        else:
            torch.cuda.manual_seed(seed)
        torch.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(seed)
        random.seed(seed)
        self.log.debug(f"Random seed set:{seed}")

    def get_rand_state(self) -> dict:
        return {
            "cuda": torch.cuda.get_rng_state_all()
            if self.conf.env.type != "single"
            else torch.cuda.get_rng_state(),
            "torch": torch.get_rng_state(),
            "numpy": np.random.get_state(),
            "random": random.getstate(),
        }

    def set_rand_state(self, state: dict) -> None:
        if self.conf.env.type != "single":
            torch.cuda.set_rng_state_all(state["cuda"])
            for i, rng_state in enumerate(state["cuda"]):
                torch.cuda.set_rng_state(rng_state)
        else:
            torch.cuda.set_rng_state(state["cuda"])
        torch.set_rng_state(state["torch"])
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.set_state(state["numpy"])
        random.setstate(state["random"])
        self.log.debug("Random state set.")

    def load_state(self, path: str, force:bool=False) -> None:
        resumed = torch.load(path)
        if resumed["scheduler"]:
            self.resume_scheduler(resumed["scheduler"])
        self.resume_optimizer(resumed["optimizer"])
        self.set_rand_state(resumed["rng_state"])
        if force or not self.conf.scheme.unit:
            self.conf.scheme.unit = resumed["unit"]
        if force or not self.conf.scheme.init_iter:
            self.conf.scheme.init_iter = resumed["timestep"]
        if force or not self.conf.loader.batch_size:
            self.conf.scheme.batch_size = resumed["batch_size"]
        if force or not self.conf.loader.total_iters:
            self.conf.scheme.total_iters = resumed["total_iters"]
        self.conf.env = resumed["env"]
        self.log.debug("Random state loaded.")

    def resume_optimizer(self, state_dict=None, path:Union[str, Path]=None) -> None:
        if state_dict:
            self.opt.load_state_dict(state_dict)
        else:
            with open(path, "rb") as f:
                self.opt.load_state_dict(torch.load(f)["optimizer"])

    def resume_scheduler(self, state_dict=None, path: Union[str, Path] = None) -> None:
        if state_dict:
            self.sch.load_state_dict(state_dict)
        else:
            with open(path, "rb") as f:
                self.sch.load_state_dict(torch.load(f)["scheduler"])

    def prepare(
        self,
        device: Union[int, str],
        log=None,
        load_data: bool = True,
        load_learner: bool = True,
    ):

        # logger
        if dist.is_initialized():
            _rank = dist.get_rank()
            log.reload(log.name + f".rank{_rank}")
            if _rank:
                log.be_inactive()
            else:
                log.load_rich_console()
        log.debug("Logger ready.")

        # data
        if load_data:
            self.course.load()
            self.loader = self.get_loader(self.conf.data_label)
            self.conf.scheme.batch_per_update = len(self.loader)
        log.debug("Data ready.")

        # learner
        if load_learner:
            self.learner.prepare()
            model = self.learner.model.to(device)
            log.debug("Model loaded.")

            if self.conf.env.type in ENV_DIST_TYPES:
                if self.conf.env.type == "FSDP":

                    from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
                    from torch.distributed.fsdp.fully_sharded_data_parallel import (
                        BackwardPrefetch,
                        CPUOffload,
                    )
                    from torch.distributed.fsdp.wrap import (
                        always_wrap_policy,
                        enable_wrap,
                        lambda_auto_wrap_policy,
                        size_based_auto_wrap_policy,
                        transformer_auto_wrap_policy,
                        wrap,
                    )

                    if self.conf.env.dist.fsdp_policy:
                        if self.conf.env.dist.fsdp_policy == "always":
                            policy = functools.partial(
                                always_wrap_policy,
                                **self.conf.env.dist.fsdp_policy_options,
                            )
                        elif self.conf.env.dist.fsdp_policy == "size":
                            policy = functools.partial(
                                size_based_auto_wrap_policy,
                                **self.conf.env.dist.fsdp_policy_options,
                            )
                        elif self.conf.env.dist.fsdp_policy == "lambda":
                            policy = functools.partial(
                                lambda_auto_wrap_policy,
                                **self.conf.env.dist.fsdp_policy_options,
                            )
                        elif self.conf.env.dist.fsdp_policy == "transformer":
                            policy = functools.partial(
                                transformer_auto_wrap_policy,
                                **self.conf.env.dist.fsdp_policy_options,
                            )
                        self.learner.model = (
                            torch.compile(FSDP(model, auto_wrap_policy=policy))
                            if self.conf.scheme.compile
                            else FSDP(
                                model,
                                auto_wrap_policy=policy,
                            )
                        )
                    else:
                        self.learner.model = (
                            torch.compile(FSDP(model))
                            if self.conf.scheme.compile
                            else FSDP(model)
                        )
                else:  # DDP
                    self.learner.model = (
                        torch.compile(DDP(model))
                        if self.conf.scheme.compile
                        else DDP(model)
                    )
                log.debug("Distributed wrapping complete.")

            else:
                if self.conf.env.type == "DP":  # DP (Data Parallel)
                    self.learner.model = (
                        torch.compile(DP(model))
                        if self.conf.scheme.compile
                        else DP(model)
                    )
                    log.debug("Data Parallel wrapping complete.")
                else:  # Single
                    self.learner.model = (
                        torch.compile(model) if self.conf.scheme.compile else model
                    )

            # Components
            if "loss" in self.conf:
                self.loss = instantiate(self.conf.loss)
                log.debug("Loss function loaded.")
            else:
                self.loss = None

            if "optimizer" in self.conf:
                self.opt = instantiate(self.conf.optimizer, self.learner.model.parameters())
                log.debug("Optimizer loaded.")
            else:
                self.opt = None

            if "scheduler" in self.conf and self.conf.scheduler.cls:
                self.sch = instantiate(self.conf.scheduler, self.opt, last_epoch = self.conf.scheme.init_iter - 1)
                log.debug("Scheduler loaded.")
            else:
                self.sch = None

            # Resuming
            if not self.learner.resume():
                log.warning("Model checkpoint not resumed.")

            _res = True
            if "resume_path" in self.conf and self.conf.resume_path is not None:
                if isinstance(self.conf.resume_path, str):
                    resume_path = Path(self.conf.resume_path)
                if resume_path.is_file():
                    self.load_state(path=resume_path)
                    if self.sch:
                        self.sch.last_epoch = self.conf.scheme.init_iter - 1
                else:
                    raise FileNotFoundError(
                        f"No checkpoint(optimizer) exists: {resume_path.absolute()}"
                    )
                log.info(f"Training states resumed from {self.conf.resume_path}.")
            elif "resume_dir" in self.conf and self.conf.resume_dir:
                if isinstance(self.conf.resume_dir, str):
                    resume_dir = Path(self.conf.resume_dir)
                if resume_dir.is_dir():
                    self.load_state(
                        path=resume_dir.joinpath(f"state_{self.conf.resume_step}.cpt")
                    )
                    if self.sch:
                        self.sch.last_epoch = self.conf.scheme.init_iter - 1
                log.info(
                    f"Training states resumed from {resume_dir} at timestep {self.conf.resume_step}."
                )
            else:
                log.warning("Training states not resumed.")
                if self.conf.scheme.seed:
                    self.rand_all(int(self.conf.scheme.seed))
                elif not self.conf.scheme.seed:
                    self.log.warning("Random seed is not set.")
                    self.seed = random.random()
                    self.rand_all(int(self.seed))
                _res = False

            if self.conf.env.type in ENV_DIST_TYPES:
                self.opt = self.get_dist_opt(
                    self.conf.env.dist.optimizer,
                    self.opt,
                    self.learner.model.parameters(),
                    self.conf.optimizer.options.lr,
                )

            return _res

        return False

    @distributed
    def predict(
        self,
        device: Union[int, str],
        global_rank: int = None,
        silent: bool = False,
        final: bool = True,
        msg: str = "Inferencing...",
    ):
        if final:
            self.prepare(device, self.log)

        ret = []
        gt = []
        y = None

        with bar("eval", total_steps=len(self.loader), msg=msg) if (
            not dist.is_initialized() or dist.get_rank() == 0
        ) else nullcontext():

            for _, batch in enumerate(self.loader):

                batch = to_tensor(batch, self.conf.scheme.dtype, device)
                if isinstance(batch, list):
                    y = batch[1].to(device)
                    batch = batch[0]
                self.learner.to(device)
                self.learner.eval()
                with torch.no_grad():
                    if self.conf.scheme.precision is None:
                        pred = self.learner.predict(batch, device, self.log)
                    else:
                        with torch.autocast(
                            torch.device(device),
                            dtype=getattr(torch, self.precision),
                        ):
                            pred = self.learner.predict(batch, device, self.log)

                if dist.is_initialized():
                    if dist.get_rank() == 0:
                        gathered = [
                            torch.zeros_like(pred) for _ in range(dist.get_world_size())
                        ]
                        dist.gather(pred, gathered)
                        ret.append(torch.cat(gathered).cpu())
                        dist.barrier()
                        if y is not None:
                            gathered = [
                                torch.zeros_like(y)
                                for _ in range(dist.get_world_size())
                            ]
                            dist.gather(y, gathered)
                            gt.append(torch.cat(gathered).cpu())
                            dist.barrier()
                    else:
                        dist.gather(pred)
                        dist.barrier()
                        if y is not None:
                            dist.gather(y)
                            dist.barrier()

                if not dist.is_initialized() or dist.get_rank() == 0:
                    advance("eval")

        if not dist.is_initialized() or dist.get_rank() == 0:
            end("eval", not final)
            ret = torch.cat(ret).squeeze()
            if len(gt):
                gt = torch.cat(gt).squeeze()
            else:
                gt = None
        else:
            ret = gt = None

        if dist.is_initialized() and final:
            dist.destroy_process_group()

        if self.conf.scheme.save_inference:
            _dir = Path(self.conf.output_dir)
            torch.save(ret, _dir.joinpath(f"{self.conf.data_label}_inference.pkl"))
            self.log.info("Saved inference result.")

        return (ret, gt)



from typing import Literal, TypeVar
from . import Logger
from omegaconf import DictConfig

L = TypeVar("L", bound=Logger)


class Manager:
    r"""Manager scary ever"""

    def __init__(self) -> None:
        pass

    def run(self):
        raise NotImplementedError("sorry dude")

    @staticmethod
    def load_componenets(package: Literal["torch", "sklearn"], conf: DictConfig):
        if package == "torch":
            return Manager.load_torch()
        elif package == "sklearn":
            return Manager.load_sklearn()
        else:
            raise ValueError("")

    @staticmethod
    def load_torch(task: Literal["fit", "eval", "infer"], conf: DictConfig, log: L):
        from ..torch.learner import instantiate_learner
        from ..torch.grader import load_grader
        from ..torch.trainer import load_trainer

        learner = instantiate_learner(conf.learner, log)
        if task == "fit":
            learner = instantiate_learner(
                conf.learner,
                log,
            )

    @staticmethod
    def load_sklearn():
        raise NotImplementedError("sorry dude")



from typing import Callable, TypeVar, Union, Literal, Sequence, Any
from ..core import Grader, Learner, Course
from .predictor import TorchPredictor, distributed
from ..utils import Logger, end
import torch.distributed as dist
from copy import deepcopy
from omegaconf import DictConfig
import os

C = TypeVar("C", bound=Course)
M = TypeVar("M", bound=Learner)
L = TypeVar("L", bound=Logger)


def load_grader(grader_conf: DictConfig, log: L, course=None, learner=None):
    grader = TorchGrader(grader_conf, logger=log, course=course, learner=learner)
    log.info("Grader instance created.")
    return grader


class TorchGrader(Grader, TorchPredictor):
    def __init__(
        self,
        conf: DictConfig,
        methods: Union[
            str, Callable[..., Any], Sequence[Union[str, Callable[..., Any]]]
        ] = None,
        options: list[dict] = None,
        logger=None,
        course: C = None,
        learner: M = None,
    ) -> None:
        self.unloaded = (methods, deepcopy(conf.scheme.metrics))
        conf.scheme.metrics = None

        super().__init__(conf, None, options, logger, course, learner)

    def load_metrics(self) -> None:
        if self.unloaded is not None:
            self.append_metric(self.unloaded[0])
            self.append_metric(self.unloaded[1])
            self.unloaded = None

    @distributed
    def eval(
        self,
        device,
        global_rank,
        silent: bool = False,
        final: bool = True,
        learner: M = None,
        course: C = None,
        timestep: int = None,
        unit: Literal["epoch", "step"] = "epoch",
        env_conf: DictConfig = None,
    ):
        if timestep is not None and unit is not None:
            if (
                self.conf.scheme.unit != unit
                or timestep % self.conf.scheme.eval_interval != 0
            ):
                return

        if not dist.is_initialized() or dist.get_rank() == 0:
            self.load_metrics()
            master_prog = True
        else:
            master_prog = False

        if self.conf.env.type == "keep":
            self.conf.env = env_conf

        if course is not None:
            self.course = course

        if learner is not None:
            self.learner = learner

        if not final:
            self.loader = self.get_loader(self.conf.data_label)
        else:
            self.prepare(device, self.log)

        if self.conf.env.type == "remote":
            raise NotImplementedError("sorry, remote grader is still in development.")
        elif self.conf.env.type == "single":
            res, gt = self.predict(
                device=self.conf.env.dev_id,
                global_rank=0,
                silent=True,
                final=final,
                msg="Evaluating...",
            )
        elif dist.is_initialized():
            res, gt = self.predict(
                device=int(os.environ["LOCAL_RANK"]),
                global_rank=dist.get_rank(),
                silent=True,
                final=final,
                msg="Evaluating...",
            )
        else:
            res, gt = self.predict(
                silent=True,
                final=final,
                msg="Evaluating...",
            )

        if master_prog:

            if gt is None:
                self.log.warning("No ground truth for evaluation.")
            elif res.shape != gt.shape:
                self.log.warning("Predictions and ground truth have different shapes.")

            eval_result = self.compute(res, gt)
            eval_result = {k: float(v) for k, v in eval_result.items()}
            if final:
                self.log.log_state(
                    eval_result, epoch=1, unit="epoch", training_state=False
                )
            else:
                self.log.log_state(
                    eval_result, epoch=timestep, unit=unit, training_state=False
                ) if unit == "epoch" else self.log.log_state(
                    eval_result, step=timestep, unit=unit, training_state=False
                )
            return eval_result



import torch
from torch.optim.lr_scheduler import LambdaLR, SequentialLR, LinearLR
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader
from typing import Callable, List, Literal, Union
from math import sin, cos, pi
import warnings
import functools

def linear_warmup_func(timestep:int, total_steps:int):
    return (timestep + 1) / total_steps

def linear_decay_func(timestep:int, total_steps:int):
    return (total_steps - timestep) / (total_steps + 1)

def cosine_warmup_func(timestep:int, total_steps:int):
    return (sin((timestep + 1) / total_steps * pi - pi / 2) + 1) / 2

def cosine_decay_func(timestep:int, total_steps:int):
    return (cos((timestep + 1) / (total_steps + 1) * pi) + 1) / 2

class WarmupScheduler(LambdaLR):
    def __init__(
        self,
        optimizer,
        warmup_steps:int,
        warmup_type:Literal["linear", "cosine"]="linear",
        last_epoch:int = -1,
        verbose:bool = False,
    ) -> None:
        self.warmup_type = warmup_type
        self.warmup_steps = warmup_steps
        if self.warmup_type == "linear":
            super().__init__(
                optimizer,
                functools.partial(linear_warmup_func, total_steps=warmup_steps),
                last_epoch,
                verbose)
        elif self.warmup_type == "cosine":
            super().__init__(
                optimizer,
                functools.partial(cosine_warmup_func, total_steps=warmup_steps),
                last_epoch,
                verbose)

class LinearDecayScheduler(LambdaLR):
    def __init__(
        self,
        optimizer,
        decay_steps:int,
        last_epoch:int = -1,
        verbose:bool = False,
    ) -> None:
        self.decay_steps = decay_steps
        super().__init__(
            optimizer,
            functools.partial(linear_decay_func, total_steps=decay_steps),
            last_epoch,
            verbose)

class CosineDecayScheduler(LambdaLR):
    def __init__(
        self,
        optimizer,
        decay_steps:int,
        last_epoch:int = -1,
        verbose:bool = False,
    ) -> None:
        self.decay_steps = decay_steps
        super().__init__(
            optimizer,
            functools.partial(cosine_decay_func, total_steps=decay_steps),
            last_epoch,
            verbose)

class LinearDecayWithWarmupScheduler(SequentialLR):
    def __init__(
        self,
        optimizer,
        total_iters:int,
        warmup_steps:int,
        warmup_type:Literal["linear", "cosine"]="linear",
        last_epoch:int = -1,
        verbose:bool = False,
    ) -> None:
        self.warmup_sch = WarmupScheduler(optimizer, warmup_steps, warmup_type, last_epoch, verbose)
        self.decay_sch = LinearDecayScheduler(optimizer, total_iters - warmup_steps, last_epoch, verbose)
        super().__init__(optimizer, [self.warmup_sch, self.decay_sch], [warmup_steps], last_epoch, verbose)

class CosineDecayWithWarmupScheduler(SequentialLR):
    def __init__(
        self,
        optimizer,
        total_iters:int,
        warmup_steps:int,
        warmup_type:Literal["linear", "cosine"]="linear",
        last_epoch:int = -1,
        verbose:bool = False,
    ) -> None:
        self.warmup_sch = WarmupScheduler(optimizer, warmup_steps, warmup_type, last_epoch, verbose)
        self.decay_sch = CosineDecayScheduler(optimizer, total_iters - warmup_steps, last_epoch, verbose)
        super().__init__(optimizer, [self.warmup_sch, self.decay_sch], [warmup_steps], last_epoch, verbose)


from typing import Sequence, Union, TypeVar, Callable, Literal
from numpy import ndarray
import sklearn.metrics as mt
from ..core import Learner, Course
from omegaconf import DictConfig

T_co = TypeVar("T_co", covariant=True)
L = TypeVar("L", bound=Learner)
C = TypeVar("C", bound=Course)

sk_metric = {
    "acc": mt.accuracy_score,
    "f1": mt.f1_score,
    "top_k_acc": mt.top_k_accuracy_score,
    "fbeta": mt.fbeta_score,
    "hamming": mt.hamming_loss,
    "hinge": mt.hinge_loss,
    "jaccard": mt.jaccard_score,
    "log": mt.log_loss,
    "precision": mt.precision_score,
    "recall": mt.recall_score,
    "roc_auc": mt.roc_auc_score,
    "pr_curve": mt.precision_recall_curve,
    "roc_curve": mt.roc_curve,
    "01": mt.zero_one_loss,
    "max": mt.max_error,
    "mae": mt.mean_absolute_error,
    "mse": mt.mean_squared_error,
    "msle": mt.mean_squared_log_error,
    "meae": mt.median_absolute_error,
}


def sk_eval(preds, gt, methods: Union[str, Sequence[str]], **kwargs):

    if isinstance(method, str):
        methods = [method]

    if not isinstance(preds, ndarray):
        preds = preds.numpy()

    if not isinstance(gt, ndarray):
        gt = gt.numpy()

    ret = []

    for method in methods:
        ret.append(sk_metric[method](gt, preds, **kwargs))

    return ret



"""
Config utilities
"""

from omegaconf import DictConfig, OmegaConf
from importlib.resources import files
from importlib.util import spec_from_file_location, module_from_spec
from importlib import import_module
import sys
import os
from pathlib import Path
from typing import Union, Sequence
from omegaconf import OmegaConf
from copy import deepcopy
from functools import partial

RESKEYS = (
    RESKEY_EVAL := "eval",
    RESKEY_LAMBDA := "ld",
    RESKEY_LAMBDA_EXP := "exp",
    RESKEY_INST := "ins",
)

RESOLVERS = {
    RESKEY_EVAL: lambda x: eval(x),
    RESKEY_LAMBDA: lambda x: lambda: eval(x),
    RESKEY_LAMBDA_EXP: lambda e: lambda x: x**e,
    RESKEY_INST: lambda p, c, o: load_class(p, c)(**o),
}

for reskey in RESKEYS:
    OmegaConf.register_new_resolver(reskey, RESOLVERS[reskey], replace=True)


def recursive_merge(*configs) -> DictConfig:

    merged = configs[0]
    if len(configs) == 1:
        return merged
    else:
        for config in configs[1:]:
            if not config:
                continue
            merged = OmegaConf.unsafe_merge(merged, config)

    return merged


def recursive_merge_by_path(*paths) -> DictConfig:
    if len(paths) == 1:
        return OmegaConf.load(paths[0])
    else:
        configs = [OmegaConf.load(path) for path in paths]
        return recursive_merge(*configs)


def load_conf(path: Union[str, Path, Sequence[Union[str, Path]]] = None) -> DictConfig:
    """
    Load yaml file
    """
    conf = OmegaConf.load(files("eniat").joinpath("config/default.yaml"))
    if path is not None:
        if isinstance(path, str) or isinstance(path, Path):
            path = [path]
        return recursive_merge_by_path(conf, path)
    else:
        return conf


def import_by_file(path: Union[str, Path], name: str = None):
    """
    import a .py file as a module by path
    """
    _spec = spec_from_file_location(
        _bn := name if name else os.path.basename(path), os.path.abspath(path)
    )
    if not _spec:
        raise FileNotFoundError(f"Can not read from the location : {path}")
    _mod = module_from_spec(_spec)
    sys.modules[_bn] = _mod
    _spec.loader.exec_module(_mod)
    return _mod, _bn


def load_class(path: Union[str, Path] = None, _class: str = None):
    """
    import a class from a module by path
    """
    if path is None:
        if _class is None:
            raise ValueError("No valid arguments")
        _cls = _class.split(".")
        try:
            return getattr(import_module(".".join(_cls[:-1])), _cls[-1])
        except:
            return getattr(
                import_module("eniat." + ".".join(_cls[:-1]), "eniat"), _cls[-1]
            )
    else:
        return getattr(import_by_file(path)[0], _class)


def instantiate(conf: DictConfig, *args, _partial: bool = False, **kwargs):
    """
    instantiate a provided DictConfig. The config must have 'cls' property
    and optionally 'path'. 'cls' is a name for the class or the whole path
    of the class (eg. torch.nn.Linear). 'path' is necessary if the class
    is defined in a .py file and
    """
    conf = deepcopy(conf)
    if not conf:
        return None
    try:
        _cls = load_class(conf.path, conf.cls)
    except:
        raise ValueError(f"Instantiation failed. Current config is not valid: {conf}")

    args = list(args)
    options = {}
    if "options" in conf and conf.options:
        options = OmegaConf.to_container(conf.options)
        args += options.pop("__args__", [])
    return (
        partial(_cls, *args, **options, **kwargs)
        if _partial
        else _cls(*args, **options, **kwargs)
    )
